package com.bergerkiller.bukkit.nolagg.meta;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;

import net.minecraft.server.ChunkProviderServer;
import net.minecraft.server.EntityItem;
import net.minecraft.server.MathHelper;
import net.minecraft.server.WorldServer;

import org.bukkit.Chunk;
import org.bukkit.Location;
import org.bukkit.World;
import org.bukkit.craftbukkit.CraftChunk;
import org.bukkit.craftbukkit.CraftWorld;
import org.bukkit.entity.Item;
import org.bukkit.event.world.ChunkUnloadEvent;
import org.bukkit.inventory.ItemStack;

import com.bergerkiller.bukkit.nolagg.ChunkOperation.Type;
import com.bergerkiller.bukkit.nolagg.ChunkSaving;
import com.bergerkiller.bukkit.nolagg.ChunkScheduler;
import com.bergerkiller.bukkit.nolagg.ItemHandler;
import com.bergerkiller.bukkit.nolagg.NLWorldListener;
import com.bergerkiller.bukkit.nolagg.NoLagg;
import com.bergerkiller.bukkit.nolagg.Util;

public class ChunkMetaData {
	public static int unloadDelay;
	public static boolean useUnloadDelay;
	
	public static ChunkMetaData get(Chunk chunk) {
		return get(chunk.getWorld(), chunk.getX(), chunk.getZ());
	}
	public static ChunkMetaData get(final World world, final int x, final int z) {
		return WorldMetaData.get(world).getChunk(x, z);
	}
	public static ChunkMetaData get(final Location location) {
		return WorldMetaData.get(location.getWorld()).getChunk(location);
	}
	
	protected ChunkMetaData(final WorldMetaData world, final int x, int z) {
		this.world = world;
		this.spawnedItems = new ArrayList<Item>();
		this.hiddenItems = new ArrayList<Item>();
		this.x = x;
		this.z = z;
		this.add();
	}
	private final WorldMetaData world;
	private final int x, z;
	private final ArrayList<Item> spawnedItems;
	private final ArrayList<Item> hiddenItems;
	
	public long lastAccessTime;
	
	public WorldMetaData getMetaWorld() {
		return this.world;
	}
	public CraftWorld getWorld() {
		return this.world.world;
	}
	public WorldServer getNativeWorld() {
		return this.world.getNative();
	}
	public net.minecraft.server.Chunk getNativeChunk() {
		return this.getChunk().getHandle();
	}
	public CraftChunk getChunk() {
		return (CraftChunk) this.getWorld().getChunkAt(this.x, this.z);
	}
	
	public int getX() {
		return this.x;
	}
	public int getZ() {
		return this.z;
	}
	
	public boolean isLoaded() {
		return this.getWorld().isChunkLoaded(this.getX(), this.getZ());
	}
	
	
	public void touch() {
		this.lastAccessTime = System.currentTimeMillis();
	}
	
	public boolean isExpired() {
		return this.isExpired(System.currentTimeMillis() - unloadDelay);
	}
	public boolean isExpired(final long time) {
		if (!useUnloadDelay) return true;
		return this.lastAccessTime <= time;
	}
	
	public boolean isPlayerNear() {
		final int bx = this.getX() << 4;
		final int bz = this.getZ() << 4;
		net.minecraft.server.Entity e;
		for (Object o : this.getNativeWorld().players) {
			e = (net.minecraft.server.Entity) o;
			if (Math.abs(MathHelper.floor(e.locX) - bx) > Util.blockView) continue;
			if (Math.abs(MathHelper.floor(e.locZ) - bz) > Util.blockView) continue;
			return true;
		}
		return false;
	}
	public boolean isUnloadable() {
		if (!this.getWorld().isAutoSave()) return false;
		if (this.getWorld().getKeepSpawnInMemory()) {
			Location spawn = this.getWorld().getSpawnLocation();
			int x = this.x - (spawn.getBlockX() >> 4);
			int z = this.x - (spawn.getBlockZ() >> 4);
			return x < -12 || x > 12 || z < -12 || z > 12;
		} else {
			return true;
		}
	}

	public String toString() {
		return "[" + this.getX() + "/" + this.getZ() + "/" + this.getWorld().getName() + "]";
	}
			
	public List<Item> getSpawnedItems() {
		return this.spawnedItems;
	}
	public int getHiddenItemCount() {
		return this.hiddenItems.size();
	}
	public void clearItems() {
		this.hiddenItems.clear();
		this.spawnedItems.clear();
	}
	public boolean removeSpawnedItem(Item item) {
		if (this.spawnedItems.remove(item)) {
			this.spawnInChunk();
			return true;
		} else {
			return false;
		}
	}
	public void spawnInChunk() {
		for (int i = this.spawnedItems.size(); i <= ItemHandler.maxItemsPerChunk && this.hiddenItems.size() > 0; i++) {
			ItemHandler.respawnItem(this.hiddenItems.remove(0));
		}
	}
	public boolean handleItemSpawn(Item item, boolean doIgnoreCheck) {
		if (ItemHandler.ignoreSpawn || !NoLagg.bufferItems || ItemHandler.maxItemsPerChunk < 0) return true;
		if (ItemHandler.maxItemsPerChunk == 0) return false;
		if (doIgnoreCheck && ItemHandler.isIgnoredItem(item)) return true;
		int currentcount = this.spawnedItems.size();
		if (currentcount > ItemHandler.maxItemsPerChunk) {
			if (ItemHandler.formStacks) {
				//can we add to an existing stack or is a new item required?
				ItemStack main = item.getItemStack();
				int amount = main.getAmount();
				int max = main.getType().getMaxStackSize();
				boolean add = true;
				for (Item hiddenitem : this.hiddenItems) {
					ItemStack hiddenstack = hiddenitem.getItemStack();
					if (hiddenstack.getType() == main.getType() && hiddenstack.getDurability() == main.getDurability()) {
						int hidamount = hiddenstack.getAmount();
						if (hidamount < max) {
							if ((hidamount + amount) <= max) {
								hiddenstack.setAmount(hidamount + amount);
								add = false;
								break;
							} else {
								amount -= max - hidamount;
								hiddenstack.setAmount(max);
							}
						}
					}
				}
				if (add) {
					main.setAmount(amount);
					this.hiddenItems.add(item);
				}
			} else {
				this.hiddenItems.add(item);
			}
			return false;
		} else {
			this.spawnedItems.add(item);
			return true;
		}
	}
		
	/*
	 * Called every so much ticks and during general events
	 */
	public void update() {
		//update spawned items
		if (NoLagg.bufferItems) {
			int i = 0;
			boolean removed = false;
			while (i < this.spawnedItems.size()) {
				if (this.spawnedItems.get(i).isDead()) {
					this.spawnedItems.remove(i);
					removed = true;
				} else {
					i++;
				}
			}
			if (removed) this.spawnInChunk();
		}
	}
	
	/*
	 * Called when this chunk meta data is first added
	 */
	@SuppressWarnings("rawtypes")
	public void add() {
		//load items
		net.minecraft.server.Chunk c = this.getNativeChunk();
		Iterator it;
		for (int i = 0; i < c.entitySlices.length; i++) {
			it = ((List) c.entitySlices[i]).iterator();
			while (it.hasNext()) {
				net.minecraft.server.Entity e = (net.minecraft.server.Entity) it.next();
				if (e instanceof EntityItem && !this.handleItemSpawn((Item) e.getBukkitEntity(), true)) {
					//remove item
					e.dead = true;
				}
			}
		}
	}
	
	/*
	 * Unloads this chunk, but does NOT remove the meta data!
	 */
	protected boolean unload() {	
		//unload directly and ignore myself
		NLWorldListener.nextIgnoreChunk = this;
		CraftChunk chunk = this.getChunk();
		if (Util.callEvent(new ChunkUnloadEvent(chunk))) {
			NLWorldListener.nextIgnoreChunk = null;
			this.unloadInternal();
			if (ChunkSaving.enabled) {
				try {
					net.minecraft.server.Chunk c = chunk.getHandle();
					if (c == null) return false;
					c.removeEntities();
					ChunkProviderServer cps = this.world.getChunkProvider();
					cps.unloadQueue.remove(c.x, c.z);
					cps.chunks.remove(c.x, c.z);
					cps.chunkList.remove(c);
					ChunkScheduler.schedule(c, Type.UNLOAD);
					return true;
				} catch (Exception ex) {
					NoLagg.log(Level.WARNING, "Failed to unload chunk " + this.toString() + ":");
					ex.printStackTrace();
					return false;
				}
			} else {
				return chunk.unload();
			}
		} else {
			NLWorldListener.nextIgnoreChunk = null;
			return false;
		}
	}
	private void unloadInternal() {
		for (Item item : this.hiddenItems) ItemHandler.respawnItem(item);
	}
	
	/*
	 * Called when this chunk meta data is being removed (unloaded)
	 */
	public void remove() {
		this.remove(true);
	}
	protected void remove(final boolean removefromworldmeta) {
		//respawn items
		this.unloadInternal();
		if (removefromworldmeta) this.world.removeChunk(this);
	}
	
	public boolean isChunk(Chunk chunk) {
		if (chunk.getWorld() != this.getWorld()) return false;
		if (chunk.getX() != this.getX()) return false;
		if (chunk.getZ() != this.getZ()) return false;
		return true;
	}

}
